package algs.hw2.solution;

import algs.hw2.AllCards;
import algs.hw2.Card;
import algs.hw2.Deck;
import algs.hw2.State;
import edu.princeton.cs.algs4.Queue;
import edu.princeton.cs.algs4.SequentialSearchST;

public class FindDeals {
	
	/**
	 * Find all deals that bring each card to the top. Since there are multiple ones, any that
	 * you find (which can be validated) are allowed.
	 */
	public void findDeals() {
		SequentialSearchST<Card, Deck> ordered = new SequentialSearchST<>();
		SequentialSearchST<Card, String> shuffles = new SequentialSearchST<>();
		
		HeinemanDeck deck = new HeinemanDeck(13);
		
		State state = new State(deck, "");
		ordered.put(deck.peekTop(), deck);
		shuffles.put(deck.peekTop(), "");
		
		Queue<State> decks = new Queue<>();
		decks.enqueue(state);
		
		while (ordered.size() < deck.size()) {
			State next = decks.dequeue();
			Card top = next.deck.peekTop();
			
			if (!ordered.contains(top)) {
				ordered.put(top, next.deck);
				shuffles.put(top, next.shuffle);
			}
			
			Deck in = next.deck.copy();
			in.in();
			decks.enqueue(new State(in, next.shuffle + "I"));
				
			Deck out = next.deck.copy();
			out.out();
			decks.enqueue(new State(out, next.shuffle + "O"));
		}
		
		for (Card c : new AllCards()) {
			System.out.println(c + "\t" + shuffles.get(c));
		}
	}
	
	public static int numberOutShuffleToReturn(int numRank) {
		Deck d = new HeinemanDeck(numRank);
		d.out();
		int numDeal = 1;
		while (!d.isInOrder()) {
			d.out();
			numDeal++;
		}
		return numDeal;
	}
	
	public static int numberInShuffleToReturn(int numRank) {
		Deck d = new HeinemanDeck(numRank);
		d.in();
		int numDeal = 1;
		while (!d.isInOrder()) {
			d.in();
			numDeal++;
		}
		return numDeal;
	}
	
	public static int numberInShuffleToReverse(int numRank) {
		Deck d = new HeinemanDeck(numRank);
		d.in();
		int numDeal = 1;
		int ct = 0;
		while (!d.isInReverseOrder() && ++ct < 1000) {
			d.in();
			numDeal++;
		}
		if (ct == 1000) return -1; else return numDeal;
	}
	
	/** 
	 * This was omitted from the homework. You can move the top card to any position in the deck
	 * if you represent the desired position (1 to 52) as a binary string, using 1-bits for in() 
	 * shuffles, and 0-bits for out() shuffles. Pretty cool, huh!
	 */
	public static void moveTopCardTo(int position) {
		Deck d = new HeinemanDeck(13);
		
		String bits = Integer.toBinaryString(position);
		for (int i = 0; i < bits.length(); i++) {
			if (bits.charAt(i) == '1') {
				d.in();
			} else {
				d.out();
			}
		}
		
		System.out.println(d.match(new Card("AC"), 10));
	}
	
	public static void main(String[] args) {
		new FindDeals().findDeals();
		
		moveTopCardTo(10);
		
		System.out.println("N\tout\tin\t#rev");
		for (int n = 1; n <= 100; n++) {
			int numOut   = numberOutShuffleToReturn(n);
			int numIn    = numberInShuffleToReturn(n);
			int numInRev = numberInShuffleToReverse(n);
			System.out.println(String.format("%d\t%d\t%d\t%d", n,numOut,numIn,numInRev));
		}
	} 
}
