package algs.hw2.solution;

import algs.hw2.*;

/**
 * Copy this class into your USERID.hw2 package and complete.
 */
public class HeinemanDeck extends Deck {
	
	/** Number of cards. */
	int N;
	
	/**
	 * Ensure that no one OUTSIDE of this class invokes the no-argument constructor. You will find
	 * it useful to have this constructor within the copy() method since it must return an accurate
	 * copy of the current Deck.
	 * 
	 * Note: if you construct a Deck instance, you must populate it with cards and not simply return
	 * an empty deck.
	 */
	protected HeinemanDeck() {
		
	}
	
	/** 
	 * Construct a playing deck with {max_rank} cards in specific order.
	 * 
	 * Once done, the linked list of card Nodes must represent a deck that looks like the following (if 
	 * {max_rank} were 3). The suites are ordered Club < Diamond < Hearts < Spades.
	 * 
	 * AC -> 2C -> 3C -> AD -> 2D -> 3D -> AH -> 2D -> 3H -> AS -> 2S -> 3S
	 * 
	 * Note your deck will have 4*{max_rank} cards once done.
	 * 
	 * Performance must be O(N) where N is max_rank.
	 */
	public HeinemanDeck(int max_rank) {
		if (max_rank < 1) { throw new IllegalArgumentException("max_rank must be at least 1"); }
		
		Suit[] suits = new Suit[] { Suit.SPADES, Suit.HEARTS, Suit.DIAMONDS, Suit.CLUBS};
		
		// fills the bag such that the bottom most card is a KING of SPADES and the top card is the ACE of CLUBS
		for (Suit suit : suits) {
			for (int rank = max_rank; rank >= Card.ACE; rank--) {
				Node node = new Node(new Card(suit, rank), first);
				if (last == null) { last = node; }
				first = node;
				this.N++;
			}
		}
	}
	
	/**
	 * Return total number of cards in deck.
	 *  
	 * Performance must be O(1)
	 */
	@Override
	public int size() {
		return this.N;
	}
	
	/**
	 * Peek to see the top card in the deck.
	 * 
	 * Performance must be O(1)
	 */
	public Card peekTop() {
		return first.card;
	}
	
	/**
	 * Peek to see the bottom card in the deck. 
	 * 
	 * Performance must be O(1)
	 */
	public Card peekBottom() {
		return last.card;
	}
	
	@Override
	public boolean match(Card c, int n) {
		Node tmp = first;
		while (n > 1) {
			tmp = tmp.next;
			n--;
		}
		
		return tmp.card.equals(c);		
	}
	
	/**
	 * Return two two cards in the deck or null if deck is empty.
	 *
	 * Performance must be O(1).
	 */
	public Pair topTwo() {
		if (N == 0) { return null; }
		
		Card one = first.card;
		Card two = first.next.card;
		
		first = first.next.next;
		N -= 2;
		if (N == 0) {
			first = last = null;
		}
		return new Pair(one, two);
	}
	
	/**
	 * Add cards to bottom of the deck.
	 * 
	 * Card 1 is added first, then Card two is added next.
	 *
	 * Performance must be O(1).
	 */
	public void addCardsToBottom(Card one, Card two) {
		if (first == null) {
			first = new Node(one);
			first.next = new Node(two);
			last = first.next;
			N = 2;
			return;
		}
		
		last.next = new Node(one);
		last.next.next = new Node(two);
		N += 2;
		last = last.next.next;
	}
	
	/** 
	 * Return a duplicate deck, with the cards in order.
	 * 
	 * Performance must be O(N) where N is the number of cards in the deck.
	 */
	public HeinemanDeck copy() {
		HeinemanDeck d = new HeinemanDeck();
		
		// make a copy of the cards into a new linked list.
		Node n = first;
		Node copy = null; // last one in the copy
		Node last = null;
		while (n != null) {
			if (d.first == null) {
				copy = new Node(n.card);
				d.first = copy;
				last = copy;
			} else {
				last.next = new Node(n.card);
				last = last.next;
			}
			
			n = n.next;
		}
		
		d.N = N;
		return d;
	}

	/** 
	 * This operation will return the first node in a linked list of 
	 * Card Nodes while MODIFYING THE CURRENT DECK TO ONLY CONTAIN THE 
	 * CARDS IN THE BOTTOM HALF OF THE DECK.
	 * 
	 * Current linked list retains only the first N/2 elements.
	 * 
	 * IF THIS METHOD IS CALLED WHEN THE DECK HAS AN ODD NUMBER OF CARDS,
	 * THEN IT MUST THROW A RuntimeException for protection.
	 * 
	 * Note that this method is never called externally, so you are in control
	 * of when you call it.
	 * 
	 * Performance must be O(N) where N is the number of cards in the deck.
	 */
	protected Node cutInHalf() {
		// start both off at same place
		Node n1 = first;
		Node n2 = first;
		
		while (n1 != null) {
			
			// Try to advance n2 by two. 
			n2 = n2.next; 
			
			// If not possible then n2 has "fallen off" the edge of list, and n1 is the "last" one of the left-side 
			if (n2 == null || n2.next == null) {
				// sever the connection between the left side (from first up until n1) and the right side (n1.next and beyond)
				Node secondHalf = n1.next;
				n1.next = null;
				Node toReturn = first;
				first = secondHalf;

				// simply return the linked list.
				this.N /= 2;
				return toReturn;
			}
			
			// we were able to advance two. Now advance n1 just one (which ensures we get to halfway point).
			n2 = n2.next;
			n1 = n1.next; 
		}
		
		// empty to begin with!
		return null;
	}
	
	/** 
	 * Conduct an out shuffle where the deck is cut in half and the TOP and BOTTOM cards remain in their place.
	 * 
	 * If the Deck contained EIGHT cards and 'first' points to the TOP of the deck (AH below) 
	 * 
	 * AC -> 2C -> AD -> 2D -> AH -> 2H -> AS -> 2S 
	 * 
	 * Then an 'out' shuffle would look like this:
	 * 
	 * LEFT:    AC    2C    AD    2D
	 * RIGHT:    | AH  | 2H  | AS  | 2S
     *           |  |  |  |  |  |  |  |
	 *           v  v  v  v  v  v  v  v
	 *          
	 * RESULT:  AC AH 2C 2H AD AS 2D 2S 
	 *  
	 * Which shows that the top and bottom cards remain the same. 
	 * 
	 * Hint: It is acceptable to create new Node objects as part of this method, if you want.
	 * 
	 * Performance must be O(N) where N is the number of cards in the deck.
	 */
	public void out() {
		// Use the available methods of BAG plus your knowledge of its STRUCTURE to complete this method.
		Node left = cutInHalf();
		Node right = first;

		Node combined = null;
		Node tail = null;
		while (left != null) {			
			// advance left and right in lock step fashion.
			if (combined == null) {    
				combined = left;
				left = left.next;
				combined.next = right;
			} else {
				tail.next = left;
				left = left.next;
				tail = tail.next;
				tail.next = right;
			}
			tail = right;
			right = right.next;
		}
		
		first = combined;
		last = tail;
		this.N *= 2;
	}

	/** 
	 * Conduct an in shuffle where the deck is cut in half and the TOP and BOTTOM cards remain in their place.
	 * 
	 * If the Deck contained EIGHT cards and 'first' points to the TOP of the deck (AH below) 
	 * 
	 * AC -> 2C -> AD -> 2D -> AH -> 2H -> AS -> 2S 
	 * 
	 * Then an 'in' shuffle would look like this:
	 * 
	 * LEFT:       AC    2C    AD    2D
	 * RIGHT:   AH  | 2H  | AS  | 2S  |
	 *           |  |  |  |  |  |  |  |
	 *           v  v  v  v  v  v  v  v
	 *          
	 * IN:      AH AC 2H 2C AS AD 2S 2D    
	 *  
	 * Which shows that the original top card is second, and the original bottom card is second from the bottom.
	 * 
	 * Performance must be O(N) where N is the number of cards in the deck.
	 */
	public void in () {
		Node left = cutInHalf();
		Node right = first;
		
		Node combined = null;
		Node tail = null;
		while (left != null) {
			
			// advance left and right in lock step fashion.
			if (combined == null) {    
				combined = right;
				right = right.next;
				combined.next = left;
				tail = left;
			} else {
				tail.next = right;
				right = right.next;
				tail = tail.next;
				tail.next = left;
				tail = tail.next;
			}
			tail = left;
			left = left.next;
		}
		first = combined;
		last = tail;
		this.N *= 2;
	}
	
	public String representation() {
		StringBuffer output = new StringBuffer();
		Node n = first;
		while (n != null) {
			output.append(n.card).append(" ");
			n = n.next;
		}
		
		return output.toString();
	}
	
	public static void produceTable() {
		for (int r = 1; r <= 13; r++ ) {
			HeinemanDeck d = new HeinemanDeck(r);
			//System.out.println(d.representation());
			String match = d.representation();
			for (int i = 0; i < 90; i++) {
				d.out();
				if (d.representation().equals(match)) {
					System.out.println(r + "\t" + i);
					break;
				}
			}
		}
	}
	
	
	public static void main(String[] args) {
		Deck deck2 = new HeinemanDeck(13);
		for (int i = 0; i < 52; i++) {
			System.out.println(deck2);
			deck2.in();
		}
	}

	@Override
	public boolean isInOrder() {
		int num = N/4;
		Node n = first;
		
		for (Suit s : Suit.values()) {
			for (int rank = 1; rank <= num; rank++) {
				if (n.card.suit != s) { return false; }
				if (n.card.rank != rank) { return false; }
				n = n.next;
			}
		}
		
		return true;
	}

	@Override
	public boolean isInReverseOrder() {
		int num = N/4;
		Node n = first;
		
		for (Suit s : new Suit[] {Suit.SPADES, Suit.HEARTS, Suit.DIAMONDS, Suit.CLUBS}) {
			for (int rank = num; rank >= 1; rank--) {
				if (n.card.suit != s) { return false; }
				if (n.card.rank != rank) { return false; }
				n = n.next;
			}
		}
		
		return true;
	}


}
